#ifndef VTKWRITER_HH
#define VTKWRITER_HH

#include "grid.hh"

#include "gw_bc.hh"
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

class VTKWriter
{
private:
  const Grid & grid_;
  std::ofstream file;

#ifdef ELLIPTIC_VTK
  class EllipticConductivity
  {
  private:
    const SpatialParameters & conductivity;
  public:

    EllipticConductivity(const SpatialParameters & _conductivity)
      : conductivity(_conductivity)
    {}
    
    inline double operator()(const size_t eid, const double )const {
      return conductivity(eid);
    }
  };
#endif

#ifdef RICHARDS_VTK
  class RichardsConductivity
  {
  private:
    typedef RichardsAssembler::SoilWaterRetention SWR;
    typedef RichardsAssembler::HydraulicConductivity HC;
    const SWR & soil_water_retention;
    const HC & hydraulic_conductivity;
  public:
    
    RichardsConductivity(const SWR & _soil_water_retention, const HC & _hydraulic_conductivity)
      : soil_water_retention(_soil_water_retention), hydraulic_conductivity(_hydraulic_conductivity)
    {
      
    }
    
    inline double operator()(const size_t eid, const double v)const {
      return hydraulic_conductivity(eid,soil_water_retention(eid,v));
    }

    inline double getSaturation(const size_t eid, const double v)const {
      return soil_water_retention(eid,v);
    }
  };
#endif

  
public:
  VTKWriter(const Grid & grid) : grid_(grid)
  {
  }


  void write_scalar(const std::string filename, 
	     const Vector & concentration)
  {
    string extfilename = filename + ".vts";
    file.open( extfilename.c_str() );

    // Generic VTK Header
    file << "<?xml version=\"1.0\"?>" << endl
	 << 
      "<VTKFile type=\"StructuredGrid\" version=\"0.1\" "
      "byte_order=\"LittleEndian\">" << endl ;

    ///////////////////////////////////////////////////////
    // Grid extent
    Vector extent = grid_.extent();
    std::vector<size_t> cells = grid_.cells();

    // if DIMENSIONS==2 add zero as vtk thinks in 3D
    if(cells.size() == 2)
      cells.push_back(0);

    file << "<StructuredGrid  WholeExtent=\"0 "
	 <<cells[0]<<" 0 "
	 <<cells[1]<<" 0 "
	 <<cells[2]<<"\">" 
	 << endl;

    file << "<Piece Extent=\"0 "
	 <<cells[0]<<" 0 "
	 <<cells[1]<<" 0 "
	 <<cells[2]<<"\">" 
	 << endl;

    //#Grid extent
    ///////////////////////////////////////////////////////

    ///////////////////////////////////////////////////////
    // Write concentration cell data
    file << "<CellData Scalars=\"Pressure\" Vectors=\"Fluxes\">" << endl;
    file << 
      "<DataArray type=\"Float64\" Name=\"Concentration\" "
      "format=\"ascii\" >"
	 << endl;

	file.setf( std::ios_base::fixed );
    for(size_t i=0; i<concentration.size(); ++i)
      file << setw(10) << setprecision(5) << concentration[i] << " ";
      
    file << "</DataArray>"<< endl;
    //#Write pressure cell data
    ///////////////////////////////////////////////////////
    file << "</CellData>" << endl;

    ///////////////////////////////////////////////////////
    // Write grid points
    file << "<Points>" << endl;
    file << 
      "<DataArray type=\"Float64\" Name=\"Coordinates\" "
      "NumberOfComponents=\"3\" format=\"ascii\">" << endl;

    Vector h(3);
    for(size_t d=0; d<Grid::dimensions; ++d)
      h[d]=extent[d]/(double)cells[d];

    const size_t cells3d = Grid::dimensions==3 ? cells[2] : 0;
    for(size_t z = 0; z<cells3d+1; ++z)
      for(size_t y = 0; y < cells[1]+1; ++y)
	for(size_t x = 0; x < cells[0]+1; ++x)
	  file << (double)x * h[0] << " " 
	       << (double)y * h[1] << " "
	       << (double)z * h[2] << " "
	       << endl;

    file << "</DataArray>" << endl << "</Points>" << endl;

    //#Write grid points
    ///////////////////////////////////////////////////////

    file << "</Piece>" << endl;
    file << "</StructuredGrid>" << endl;
    file << "</VTKFile>" << endl;
    file.close();

  }

#ifdef ELLIPTIC_VTK
  void write_groundwater(const std::string filename, 
	     const Vector & pressure, 
	     const SpatialParameters & conductivity,
	     BoundaryConditions & boundary_conditions)
  {
    EllipticConductivity gwc(conductivity);
    write_header(filename);
    write_pressure(pressure);
    write_flux(pressure,gwc,boundary_conditions);
    write_conductivity(pressure,gwc);
    write_tail();
  }
#endif

#ifdef RICHARDS_VTK
  void write_richards(const std::string filename, 
		      const Vector & pressure, 
		      const RichardsAssembler::SoilWaterRetention swr,
		      const RichardsAssembler::HydraulicConductivity hc,
		      BoundaryConditions & boundary_conditions)
  {
    RichardsConductivity rc(swr,hc);
    write_header(filename);
    write_pressure(pressure);
    write_flux(pressure,rc,boundary_conditions);
    write_conductivity(pressure,rc);
    write_saturation(pressure,rc);
    write_tail();
  }
#endif


  template<class C>
  void write_flux(const Vector & pressure, 
		  const C & conductivity,
		  BoundaryConditions & boundary_conditions)
  {
    ///////////////////////////////////////////////////////
    // Write flux cell data
    file << 
      "<DataArray type=\"Float64\" Name=\"Fluxes\" "
      "NumberOfComponents=\"3\" format=\"ascii\">"
	 << endl;

    ElementIterator it = grid_.begin();
    ElementIterator eit = grid_.end();
    for(;it!=eit;++it){
      const size_t e_id = it.id();
      
      Vector flux(3,0.0);

      FaceIterator fit = (*it).begin();
      FaceIterator efit = (*it).end();
      for(;fit!=efit;++fit){

	const unsigned int normal_component = (*fit).normal_component();
	const int normal_direction = (*fit).normal_direction();

	if(!(*fit).is_boundary()){
	  const size_t f_id = fit.out_id();

	  const double k_e = conductivity(e_id,pressure[e_id]);
	  const double k_f = conductivity(f_id,pressure[f_id]);
	  const double k = 2. / (1./k_e + 1./k_f);

	  const double J = - (double)normal_direction * 
	    (pressure[f_id]-pressure[e_id])
	    / (*fit).dist_to_barycenter() *  0.5 * k;

	  flux[normal_component] += 0.5 * J;

	  if(normal_component==(unsigned int)GRAVITY_AXES){
	    const double G = k;
	    flux[normal_component] += -0.5 * G;
	  }


	}
	else{
	  Vector barycenter = (*it).barycenter();
	  const double h_2 = (*it).extent()[normal_component] * 0.5;
	  barycenter[normal_component] += (double)normal_direction * h_2;

	  // if dirichlet condition
	  double dirichlet_value=0;
	  if(boundary_conditions.is_dirichlet(barycenter, &dirichlet_value)){
	    const double k_e = conductivity(e_id,pressure[e_id]);
	    const double k = k_e;

	    const double J = - (double)normal_direction * (dirichlet_value - pressure[e_id])
	      / h_2 * k;

	    flux[normal_component] += J *0.5;

	    if(normal_component==(unsigned int)GRAVITY_AXES){
	      const double G = k;
	      flux[normal_component] += -0.5 * G;
	    }

	  }
	  // if neumann condition
	  else{
	    double neumann_value = boundary_conditions.get_neumann(barycenter);
	    flux[normal_component] += neumann_value *0.5;
	  }
	}
      }

      file << flux[0] << " " << flux[1] << " " << flux[2] << " ";
    }
    file << "</DataArray>"<< endl;
    //#Write flux cell data
    ///////////////////////////////////////////////////////

  }

  void write_header(const std::string filename)
  {
    string extfilename = filename + ".vts";
    file.open( extfilename.c_str() );

    // Generic VTK Header
    file << "<?xml version=\"1.0\"?>" << endl
	 << 
      "<VTKFile type=\"StructuredGrid\" version=\"0.1\" "
      "byte_order=\"LittleEndian\">" << endl ;

    ///////////////////////////////////////////////////////
    // Grid extent
    Vector extent = grid_.extent();
    std::vector<size_t> cells = grid_.cells();

    // if DIMENSIONS==2 add zero as vtk thinks in 3D
    if(cells.size() == 2)
      cells.push_back(0);

    file << "<StructuredGrid  WholeExtent=\"0 "
	 <<cells[0]<<" 0 "
	 <<cells[1]<<" 0 "
	 <<cells[2]<<"\">" 
	 << endl;

    file << "<Piece Extent=\"0 "
	 <<cells[0]<<" 0 "
	 <<cells[1]<<" 0 "
	 <<cells[2]<<"\">" 
	 << endl;

    //#Grid extent
    ///////////////////////////////////////////////////////

    ///////////////////////////////////////////////////////
    // Write pressure cell data
    file << "<CellData Scalars=\"Pressure\" Vectors=\"Fluxes\">" << endl;
  }

  template<class C>
  void write_conductivity(const Vector & pressure, 
		  const C & conductivity)
  {
    ///////////////////////////////////////////////////////
    // Write conductivity cell data
    file << 
      "<DataArray type=\"Float64\" Name=\"Sat. Conductivity\" "
      "format=\"ascii\">"
	 << endl;

    for(size_t i=0; i<pressure.size(); ++i)
      file << conductivity(i,0) << " ";
      
    file << "</DataArray>"<< endl;

    //#Write pressure cell data
    ///////////////////////////////////////////////////////
  }

  template<class C>
  void write_saturation(const Vector & pressure, 
		  const C & conductivity)
  {
    ///////////////////////////////////////////////////////
    // Write conductivity cell data
    file << 
      "<DataArray type=\"Float64\" Name=\"Saturation\" "
      "format=\"ascii\">"
	 << endl;

    for(size_t i=0; i<pressure.size(); ++i){
      file << conductivity.getSaturation(i,pressure[i]) << " ";
    }
      
    file << "</DataArray>"<< endl;

    //#Write pressure cell data
    ///////////////////////////////////////////////////////
  }


  void write_pressure(const Vector & pressure)
  {
    file << 
      "<DataArray type=\"Float64\" Name=\"Pressure\" "
      "format=\"ascii\" >"
	 << endl;

    for(size_t i=0; i<pressure.size(); ++i)
      file << pressure[i] << " ";
      
    file << "</DataArray>"<< endl;
  }


  void write_tail()
  {
    file << "</CellData>" << endl;

    ///////////////////////////////////////////////////////
    // Write grid points
    file << "<Points>" << endl;
    file << 
      "<DataArray type=\"Float64\" Name=\"Coordinates\" "
      "NumberOfComponents=\"3\" format=\"ascii\">" << endl;

    Vector extent = grid_.extent();
    std::vector<size_t> cells = grid_.cells();

    Vector h(3);
    for(size_t d=0; d<Grid::dimensions; ++d)
      h[d]=extent[d]/(double)cells[d];

    const size_t cells3d = Grid::dimensions==3 ? cells[2] : 0;
    for(size_t z = 0; z<cells3d+1; ++z)
      for(size_t y = 0; y < cells[1]+1; ++y)
	for(size_t x = 0; x < cells[0]+1; ++x)
	  file << (double)x * h[0] << " " 
	       << (double)y * h[1] << " "
	       << (double)z * h[2] << " "
	       << endl;

    file << "</DataArray>" << endl << "</Points>" << endl;

    //#Write grid points
    ///////////////////////////////////////////////////////

    file << "</Piece>" << endl;
    file << "</StructuredGrid>" << endl;
    file << "</VTKFile>" << endl;
    file.close();
  }
};

#endif 
